import numpy as np
import math

def cubic_solve_multi(a0, b0, c0, d0, all_roots=True):
    # Cubic coefficients
    a, b, c = b0 / a0, c0 / a0, d0 / a0

    # Some repeating constants and variables
    third = 1.0 / 3.0
    a_third = a * third
    a_third_square = a_third * a_third
    sqr3 = math.sqrt(3)

    # Additional intermediate variables
    f = third * b - a_third_square
    g = a_third * (2 * a_third_square - b) + c
    h = 0.25 * g * g + f * f * f

    # Masks for different combinations of roots
    m1 = (f == 0) & (g == 0) & (h == 0)     # roots are real and equal
    m2 = (~m1) & (h <= 0)                   # roots are real and distinct
    m3 = (~m1) & (~m2)                      # one real root and two complex

    def cubic_root(x):
        ''' Compute cubic root of a number while maintaining its sign
        '''
        root = np.zeros_like(x)
        positive = (x >= 0)
        negative = ~positive
        root[positive] = x[positive] ** third
        root[negative] = -(-x[negative]) ** third
        return root

    def roots_all_real_equal(c):
        ''' Compute cubic roots if all roots are real and equal
        '''
        r1 = -cubic_root(c)
        if all_roots:
            return r1, r1, r1
        else:
            return r1

    def roots_all_real_distinct(a13, f, g, h):
        ''' Compute cubic roots if all roots are real and distinct
        '''
        j = np.sqrt(-f)
        k = np.arccos(-0.5 * g / (j * j * j))
        m = np.cos(third * k)
        r1 = 2 * j * m - a13
        if all_roots:
            n = sqr3 * np.sin(third * k)
            r2 = -j * (m + n) - a13
            r3 = -j * (m - n) - a13
            return r1, r2, r3
        else:
            return r1

    def roots_one_real(a13, g, h):
        ''' Compute cubic roots if one root is real and other two are complex
        '''
        sqrt_h = np.sqrt(h)
        S = cubic_root(-0.5 * g + sqrt_h)
        U = cubic_root(-0.5 * g - sqrt_h)
        S_plus_U = S + U
        r1 = S_plus_U - a13
        if all_roots:
            S_minus_U = S - U
            r2 = -0.5 * S_plus_U - a13 + S_minus_U * sqr3 * 0.5j
            r3 = -0.5 * S_plus_U - a13 - S_minus_U * sqr3 * 0.5j
            return r1, r2, r3
        else:
            return r1

    # Compute roots
    if all_roots:
        roots = np.zeros((3, len(a))).astype(complex)
        roots[:, m1] = roots_all_real_equal(c[m1])
        roots[:, m2] = roots_all_real_distinct(a_third[m2], f[m2], g[m2], h[m2])
        roots[:, m3] = roots_one_real(a_third[m3], g[m3], h[m3])
    else:
        roots = np.zeros(len(a))
        roots[m1] = roots_all_real_equal(c[m1])
        roots[m2] = roots_all_real_distinct(a_third[m2], f[m2], g[m2], h[m2])
        roots[m3] = roots_one_real(a_third[m3], g[m3], h[m3])

    return roots